from typeguard import typechecked
from ..models.error_models import GuideCountErrorType, GuideCountError
from ..models.quality_control_models import SingleInferenceQualityControlResult, MatchSetSingleInferenceQualityControlResult, QualityControlResult
from ..models.mapping_models import GeneralMappingInferenceDict
from typing import DefaultDict, Tuple, Optional, Union
from collections import defaultdict
from ..processing.crispr_editing_processing import get_non_error_dict

# -----------------------------
# QC helpers
# -----------------------------

def iter_all_obs_values(d, contains_sample_barcode: bool):
    """
    Iterate over all observed values regardless of nesting.
    If sample barcodes are present, d is nested: {sample: {guide: InferenceResult}}
    Otherwise, d is flat: {guide: InferenceResult}
    """
    if contains_sample_barcode:
        for sample_dict in d.values():
            for obs_value in sample_dict.values():
                yield obs_value
    else:
        for obs_value in d.values():
            yield obs_value


def get_umi_noncollapsed_counts(d, contains_guide_umi: bool, contains_sample_barcode: bool):
    return sum(
        sum(v.observed_value.values()) if contains_guide_umi else v.observed_value
        for v in iter_all_obs_values(d, contains_sample_barcode)
    )


def get_umi_collapsed_counts(d, contains_guide_umi: bool, contains_sample_barcode: bool):
    return sum(
        len(v.observed_value) if contains_guide_umi else v.observed_value
        for v in iter_all_obs_values(d, contains_sample_barcode)
    )


# -----------------------------
# Update num counts in QC results
# -----------------------------

def set_num_non_error_counts(qc_result, counts_inferred_dict, contains_guide_umi: bool, contains_sample_barcode: bool):
    if contains_guide_umi:
        qc_result.num_non_error_umi_noncollapsed_counts = get_umi_noncollapsed_counts(
            counts_inferred_dict, contains_guide_umi=True, contains_sample_barcode=contains_sample_barcode
        )
        qc_result.num_non_error_umi_collapsed_counts = get_umi_collapsed_counts(
            counts_inferred_dict, contains_guide_umi=True, contains_sample_barcode=contains_sample_barcode
        )
    else:
        qc_result.num_non_error_counts = get_umi_noncollapsed_counts(
            counts_inferred_dict, contains_guide_umi=False, contains_sample_barcode=contains_sample_barcode
        )
    return qc_result


def set_num_guide_count_error_types(qc_result, attribute_name: str, observed_guide_reporter_umi_counts_inferred, contains_guide_umi: bool, contains_sample_barcode: bool):
    guide_count_error_type_umi_noncollapsed_count: DefaultDict = defaultdict(int)
    guide_count_error_type_umi_collapsed_count: DefaultDict = defaultdict(int)
    guide_count_error_type_count: DefaultDict = defaultdict(int)

    for obs_value in iter_all_obs_values(observed_guide_reporter_umi_counts_inferred, contains_sample_barcode):
        attr = getattr(obs_value.inferred_value, attribute_name, None)
        error = getattr(attr, "error", None)

        # Only count errors, skip NO_ERROR
        if error is not None and error.guide_count_error_type.name != "NO_ERROR":
            if contains_guide_umi:
                guide_count_error_type_umi_noncollapsed_count[error.guide_count_error_type] += sum(obs_value.observed_value.values())
                guide_count_error_type_umi_collapsed_count[error.guide_count_error_type] += len(obs_value.observed_value)
            else:
                guide_count_error_type_count[error.guide_count_error_type] += obs_value.observed_value

    qc_result.guide_count_error_type_umi_noncollapsed_count = guide_count_error_type_umi_noncollapsed_count
    qc_result.guide_count_error_type_umi_collapsed_count = guide_count_error_type_umi_collapsed_count
    qc_result.guide_count_error_type_count = guide_count_error_type_count
    return qc_result


# -----------------------------
# Main QC
# -----------------------------

def perform_counts_quality_control(
    observed_guide_reporter_umi_counts_inferred,
    contains_guide_umi: bool,
    contains_guide_surrogate: bool,
    contains_guide_barcode: bool,
    contains_sample_barcode: bool,
):
    qc_result = QualityControlResult()

    # Helper to process a single attribute
    def set_match_set_single_inference_quality_control_results(attribute_name: str):
        qc = MatchSetSingleInferenceQualityControlResult()
        qc.non_error_dict = get_non_error_dict(observed_guide_reporter_umi_counts_inferred, attribute_name)
        qc = set_num_non_error_counts(qc, qc.non_error_dict, contains_guide_umi, contains_sample_barcode)
        qc = set_num_guide_count_error_types(qc, attribute_name, observed_guide_reporter_umi_counts_inferred, contains_guide_umi, contains_sample_barcode)
        return qc

    # Assign QC results
    qc_result.protospacer_match = set_match_set_single_inference_quality_control_results("protospacer_match")
    if contains_guide_barcode:
        qc_result.protospacer_match_barcode_match = set_match_set_single_inference_quality_control_results("protospacer_match_barcode_match")
    if contains_guide_surrogate:
        qc_result.surrogate_match = set_match_set_single_inference_quality_control_results("surrogate_match")
    if contains_sample_barcode:
        qc_result.sample_barcode_match = set_match_set_single_inference_quality_control_results("sample_barcode_match")

    # Shared totals
    if contains_guide_umi:
        qc_result.total_umi_noncollapsed_counts = get_umi_noncollapsed_counts(
            observed_guide_reporter_umi_counts_inferred, contains_guide_umi, contains_sample_barcode
        )
        qc_result.total_umi_collapsed_counts = get_umi_collapsed_counts(
            observed_guide_reporter_umi_counts_inferred, contains_guide_umi, contains_sample_barcode
        )

    return qc_result


from typeguard import typechecked
from typing import Union, List, Mapping, Tuple, Optional, Any, DefaultDict, Dict
from typing import Counter as CounterType
from ..models.mapping_models import GeneralMappingInferenceDict, GeneralMatchCountDict, GeneralMismatchCountDict, WhitelistReporterObservedSequenceMapping

from collections import Counter
from collections import defaultdict
from ..models.mapping_models import MatchSetWhitelistReporterCounterSeriesResults, CompleteInferenceMatchResult, MatchSetSingleInferenceMatchResult, SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults, SurrogateProtospacerMismatchSingleInferenceMatchResult, AllMatchSetWhitelistReporterCounterSeriesResults, InferenceResult
from .crispr_editing_processing import check_match_result_non_error, get_non_error_dict
import pandas as pd

@typechecked
def helper_get_observed_values_given_whitelist_value(whitelist_sequence_list: List[Tuple[str, Optional[str], Optional[str]]], observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict, attribute_name:str, contains_umi: bool, ambiguous_accepted:bool = True) -> WhitelistReporterObservedSequenceMapping:
    """
    Provides the set of observed sequences for a given whitelist sequence

    Args:
        whitelist_sequence_list (List[Tuple[str, Optional[str], Optional[str]]]): List of whitelist sequence to retrieve observed sequences
        observed_guide_reporter_umi_counts_inferred (GeneralMappingInferenceDict): Datastructure contain the observed sequence to mapped whitelist sequence
        attribute_name (str): string specifying white mapping match type to consider.
        contains_umi (bool): specifying whether UMI is used
        ambiguous_accepted (bool): specifying whether to consider ambiguous mapping

    Returns:
        WhitelistReporterObservedSequenceMapping: dictionary mapping between the whitelist sequence and the list of observed sequence. Each observed sequence has the corresponding count Union[int, Dict [ str, int ]] (either int if no UMI, or Dict[str, int] for UMI-collapsed and non-collpased count)
    """
    whitelist_sequence_mapping_list: WhitelistReporterObservedSequenceMapping = defaultdict(list)


    # Iterate through the inference results (will attempt to do retrieval for all requested whitelist sequences at once for optimization)
    inferred_value_results: InferenceResult
    observed_sequence: Tuple[str, Optional[str], Optional[str]]
    for observed_sequence, inferred_value_results in get_non_error_dict(observed_guide_reporter_umi_counts_inferred, attribute_name).items():
        #
        #   Get the relevant attributes
        #
        observed_value_counts: Union[int, CounterType[Optional[str]]] = inferred_value_results.observed_value # Read count of the observed sequence
        inferred_value_result: CompleteInferenceMatchResult =  inferred_value_results.inferred_value  # Whitelist inference results
        match_set_single_inference_match_result : Optional[MatchSetSingleInferenceMatchResult] = getattr(inferred_value_result, attribute_name) # Get inference result for specific mapping strategy
        assert match_set_single_inference_match_result is not None, "match_set_single_inference_match_result should not be none since this is from the non error list. Developer error."
        
        matches: pd.DataFrame = match_set_single_inference_match_result.value.matches # Get the list of ambiguous
        
        if not matches.empty:
            
            # Skip observed sequence if there are multiple matches when not accepting ambiguous mapping 
            if (ambiguous_accepted is False) and (matches.shape[0] > 1):
                continue
            
            
            for whitelist_reporter_series in matches.iterrows(): 
                # UMI-BASED COUNTING
                dict_index = tuple(whitelist_reporter_series[1])
                if dict_index in whitelist_sequence_list: # If the match is in the requested whitelist sequences, proceed
                    # Get match info to add to result
                    all_match_sequences = [tuple(whitelist_reporter_series[1]) for whitelist_reporter_series in matches.iterrows()]
                    total_match_counts = len(all_match_sequences)

                    if contains_umi:
                        assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"

                        # Calculate both UMI noncollapsed and collapsed count
                        observed_sequence_count = {
                            "umi_noncollapsed_count": sum(observed_value_counts.values()),
                            "umi_collapsed_count": len(observed_value_counts.values())
                        }

                        # Add observed sequence and count to mapping
                        whitelist_sequence_mapping_list[dict_index].append( (observed_sequence, observed_sequence_count, total_match_counts, all_match_sequences) )
                    
                    # STANDARD NON-UMI BASED COUNTING
                    else:
                        assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"
                        
                        observed_sequence_count = observed_value_counts
                        
                        # Add observed sequence and count to mapping
                        whitelist_sequence_mapping_list[dict_index].append( (observed_sequence, observed_sequence_count, total_match_counts, all_match_sequences) )

    return whitelist_sequence_mapping_list


#
#   GET THE WHITELIST COUNT PANDAS SERIES
#
# HELPER FUNCTION GETS COUNTS FOR THE THE MATCHES - defined in-function to reduce arguments being passed (NOTE: There is some duplicate code with mismatch counts function - keep in mind if making modifications)
@typechecked
def get_matchset_counterseries(
    observed_guide_reporter_umi_counts_inferred: Union[GeneralMappingInferenceDict, DefaultDict[str, GeneralMappingInferenceDict]],
    whitelist_guide_reporter_df: pd.DataFrame,
    contains_guide_umi: bool,
    contains_sample_barcode: bool,
    attribute_name: str
) -> MatchSetWhitelistReporterCounterSeriesResults:
    
    #
    #   DEFINE THE DEFAULTDICTS FOR COUNTING
    #   Key is the guide, value is the count
    #
    if contains_sample_barcode:
        ambiguous_ignored_umi_noncollapsed_counterdict  = defaultdict(lambda: defaultdict(int))
        ambiguous_ignored_umi_collapsed_counterdict     = defaultdict(lambda: defaultdict(int))
        ambiguous_ignored_counterdict                   = defaultdict(lambda: defaultdict(int))

        ambiguous_accepted_umi_noncollapsed_counterdict = defaultdict(lambda: defaultdict(int))
        ambiguous_accepted_umi_collapsed_counterdict    = defaultdict(lambda: defaultdict(int))
        ambiguous_accepted_counterdict                  = defaultdict(lambda: defaultdict(int))

        ambiguous_spread_umi_noncollapsed_counterdict   = defaultdict(lambda: defaultdict(float))
        ambiguous_spread_umi_collapsed_counterdict      = defaultdict(lambda: defaultdict(float))
        ambiguous_spread_counterdict                    = defaultdict(lambda: defaultdict(float))
    else:
        ambiguous_ignored_umi_noncollapsed_counterdict  = defaultdict(int)
        ambiguous_ignored_umi_collapsed_counterdict     = defaultdict(int)
        ambiguous_ignored_counterdict                   = defaultdict(int)

        ambiguous_accepted_umi_noncollapsed_counterdict = defaultdict(int)
        ambiguous_accepted_umi_collapsed_counterdict    = defaultdict(int)
        ambiguous_accepted_counterdict                  = defaultdict(int)

        ambiguous_spread_umi_noncollapsed_counterdict   = defaultdict(float)
        ambiguous_spread_umi_collapsed_counterdict      = defaultdict(float)
        ambiguous_spread_counterdict                    = defaultdict(float)


    #
    # ITERATE THROUGH THE NON-ERROR INFERRED RESULTS AND FILL THE COUNTS
    #
    def process_sample(observed_dict, cell_barcode=None):
        inferred_value_results: InferenceResult
        for inferred_value_results in get_non_error_dict(observed_dict, attribute_name).values():
            #
            #   Get the relevant attributes
            #
            observed_value_counts: Union[int, CounterType[Optional[str]]] = inferred_value_results.observed_value
            inferred_value_result: CompleteInferenceMatchResult =  inferred_value_results.inferred_value 
            match_set_single_inference_match_result : Optional[MatchSetSingleInferenceMatchResult] = getattr(inferred_value_result, attribute_name)
            assert match_set_single_inference_match_result is not None, "match_set_single_inference_match_result should not be none since this is from the non error list. Developer error."
            
            matches: pd.DataFrame = match_set_single_inference_match_result.value.matches
            if not matches.empty:
                # ITERATE THROUGH MATCHE(S) TO PERFORM COUNTS
                for whitelist_reporter_series in matches.iterrows(): 
                    dict_index = tuple(whitelist_reporter_series[1])

                    # Helper for incrementing either flat or nested dicts
                    def add_count(counterdict, value, spread=False):
                        if contains_sample_barcode:
                            if spread:
                                counterdict[cell_barcode][dict_index] += float(value)
                            else:
                                counterdict[cell_barcode][dict_index] += value
                        else:
                            if spread:
                                counterdict[dict_index] += float(value)
                            else:
                                counterdict[dict_index] += value

                    # UMI-BASED COUNTING
                    if contains_guide_umi:
                        assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"
                        total_reads = sum(observed_value_counts.values())
                        collapsed = len(observed_value_counts.values())
                        num_matches = float(matches.shape[0])

                        add_count(ambiguous_accepted_umi_noncollapsed_counterdict, total_reads)
                        add_count(ambiguous_accepted_umi_collapsed_counterdict, collapsed)
                        add_count(ambiguous_spread_umi_noncollapsed_counterdict, total_reads / num_matches, spread=True)
                        add_count(ambiguous_spread_umi_collapsed_counterdict, collapsed / num_matches, spread=True)

                        # If there is no ambiguous matches, then add to ambiguous_ignored counter
                        if matches.shape[0] == 1:
                            add_count(ambiguous_ignored_umi_noncollapsed_counterdict, total_reads)
                            add_count(ambiguous_ignored_umi_collapsed_counterdict, collapsed)

                    # STANDARD NON-UMI BASED COUNTING
                    else:
                        assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"
                        num_matches = float(matches.shape[0])
                        add_count(ambiguous_accepted_counterdict, observed_value_counts)
                        add_count(ambiguous_spread_counterdict, observed_value_counts / num_matches, spread=True)
                        
                        # If there is no ambiguous matches, then add to ambiguous_ignored counter
                        if matches.shape[0] == 1:
                            add_count(ambiguous_ignored_counterdict, observed_value_counts)

    # Process all samples
    if contains_sample_barcode:
        for cell_barcode, observed_dict in observed_guide_reporter_umi_counts_inferred.items():
            process_sample(observed_dict, cell_barcode)
    else:
        process_sample(observed_guide_reporter_umi_counts_inferred)


    # Helper function that converts defaultdict to series
    def create_counterseries(counterdict):
        if contains_sample_barcode:
            # Flatten nested dict
            records = [
                (cell_barcode, *key_tuple, value)
                for cell_barcode, inner in counterdict.items()
                for key_tuple, value in inner.items()
            ]
            columns = ["CellBarcode"] + list(whitelist_guide_reporter_df.columns) + ["value"]
            df = pd.DataFrame.from_records(records, columns=columns)
            return df.set_index(["CellBarcode"] + list(whitelist_guide_reporter_df.columns))["value"]
        else:
            counterseries: pd.Series = whitelist_guide_reporter_df.apply(lambda reporter: counterdict[tuple(reporter)], axis=1)
            counterseries.index = pd.MultiIndex.from_frame(whitelist_guide_reporter_df)
            return counterseries


    #
    #   CONVERT THE COUNT DICTS INTO PANDAS SERIES, since this is a more ideal structure.
    #
    match_set_whitelist_reporter_counter_series_results = MatchSetWhitelistReporterCounterSeriesResults()
    match_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_noncollapsed_counterseries = create_counterseries(ambiguous_ignored_umi_noncollapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_collapsed_counterseries = create_counterseries(ambiguous_ignored_umi_collapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_ignored_counterseries = create_counterseries(ambiguous_ignored_counterdict)

    match_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_noncollapsed_counterseries = create_counterseries(ambiguous_accepted_umi_noncollapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_collapsed_counterseries = create_counterseries(ambiguous_accepted_umi_collapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_accepted_counterseries = create_counterseries(ambiguous_accepted_counterdict)

    match_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_noncollapsed_counterseries = create_counterseries(ambiguous_spread_umi_noncollapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_collapsed_counterseries = create_counterseries(ambiguous_spread_umi_collapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_spread_counterseries = create_counterseries(ambiguous_spread_counterdict)

    return match_set_whitelist_reporter_counter_series_results



#
# HELPER FUNCTION GETS COUNTS FOR THE THE MISMATCHES - defined in-function to reduce arguments being passed (NOTE: There is some duplicate code with match counts function - keep in mind if making modifications)
#
@typechecked
def get_mismatchset_counterseries(observed_guide_reporter_umi_counts_inferred: Union[GeneralMappingInferenceDict, DefaultDict[str, GeneralMappingInferenceDict]] , whitelist_guide_reporter_df: pd.DataFrame, contains_guide_umi: bool, contains_sample_barcode: bool, attribute_name: str) -> SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults:
    #
    #   DEFINE THE DEFAULTDICTS FOR COUNTING
    #

    if contains_sample_barcode:
        ambiguous_ignored_umi_noncollapsed_match_counterdict  = defaultdict(lambda: defaultdict(int))
        ambiguous_ignored_umi_collapsed_match_counterdict  = defaultdict(lambda: defaultdict(int))
        ambiguous_ignored_match_counterdict  = defaultdict(lambda: defaultdict(int))

        ambiguous_accepted_umi_noncollapsed_match_counterdict = defaultdict(lambda: defaultdict(int))
        ambiguous_accepted_umi_collapsed_match_counterdict  = defaultdict(lambda: defaultdict(int))
        ambiguous_accepted_match_counterdict  = defaultdict(lambda: defaultdict(int))

        ambiguous_spread_umi_noncollapsed_match_counterdict  = defaultdict(lambda: defaultdict(float))
        ambiguous_spread_umi_collapsed_match_counterdict  = defaultdict(lambda: defaultdict(float))
        ambiguous_spread_match_counterdict  = defaultdict(lambda: defaultdict(float))

        # MISMATCH counters (keys are PAIRS of indices, representing the protospacer and surrogate match separately)
        ambiguous_ignored_umi_noncollapsed_mismatch_counterdict  = defaultdict(lambda: defaultdict(int))
        ambiguous_ignored_umi_collapsed_mismatch_counterdict  = defaultdict(lambda: defaultdict(int))
        ambiguous_ignored_mismatch_counterdict = defaultdict(lambda: defaultdict(int))

        ambiguous_accepted_umi_noncollapsed_mismatch_counterdict  =defaultdict(lambda:  defaultdict(int))
        ambiguous_accepted_umi_collapsed_mismatch_counterdict = defaultdict(lambda: defaultdict(int))
        ambiguous_accepted_mismatch_counterdict  = defaultdict(lambda: defaultdict(int))

        ambiguous_spread_umi_noncollapsed_mismatch_counterdict  = defaultdict(lambda: defaultdict(float))
        ambiguous_spread_umi_collapsed_mismatch_counterdict  = defaultdict(lambda: defaultdict(float))
        ambiguous_spread_mismatch_counterdict  = defaultdict(lambda: defaultdict(float))
    else:
        # MATCH counters
        ambiguous_ignored_umi_noncollapsed_match_counterdict  = defaultdict(int)
        ambiguous_ignored_umi_collapsed_match_counterdict  = defaultdict(int)
        ambiguous_ignored_match_counterdict  = defaultdict(int)

        ambiguous_accepted_umi_noncollapsed_match_counterdict  = defaultdict(int)
        ambiguous_accepted_umi_collapsed_match_counterdict  = defaultdict(int)
        ambiguous_accepted_match_counterdict  = defaultdict(int)

        ambiguous_spread_umi_noncollapsed_match_counterdict  = defaultdict(float)
        ambiguous_spread_umi_collapsed_match_counterdict = defaultdict(float)
        ambiguous_spread_match_counterdict  = defaultdict(float)

        # MISMATCH counters (keys are PAIRS of indices, representing the protospacer and surrogate match separately)
        ambiguous_ignored_umi_noncollapsed_mismatch_counterdict  = defaultdict(int)
        ambiguous_ignored_umi_collapsed_mismatch_counterdict  = defaultdict(int)
        ambiguous_ignored_mismatch_counterdict  = defaultdict(int)

        ambiguous_accepted_umi_noncollapsed_mismatch_counterdict   = defaultdict(int)
        ambiguous_accepted_umi_collapsed_mismatch_counterdict = defaultdict(int)
        ambiguous_accepted_mismatch_counterdict  = defaultdict(int)

        ambiguous_spread_umi_noncollapsed_mismatch_counterdict  = defaultdict(float)
        ambiguous_spread_umi_collapsed_mismatch_counterdict = defaultdict(float)
        ambiguous_spread_mismatch_counterdict = defaultdict(float)



    #
    # ITERATE THROUGH THE NON-ERROR INFERRED RESULTS (NOTE: If only one of the protospacer or surrogate matches but not the other, this is treated as an error, and will NOT be counted or considered (even in the single match series). For those counts, should just use the protospacer-only or surrogate-only match results)
    #
    def process_sample(observed_dict, cell_barcode=None):
        inferred_value_results: InferenceResult
        for inferred_value_results in get_non_error_dict(observed_dict, attribute_name).values():
            #
            #   Get the relevant attributes
            #
            observed_value_counts: Union[int, CounterType[Optional[str]]] = inferred_value_results.observed_value
            inferred_value_result: CompleteInferenceMatchResult = inferred_value_results.inferred_value
            surrogate_protospacer_mismatch_single_inference_match_result: Optional[SurrogateProtospacerMismatchSingleInferenceMatchResult] = getattr(inferred_value_result, attribute_name)
            assert surrogate_protospacer_mismatch_single_inference_match_result is not None, "surrogate_protospacer_mismatch_single_inference_match_result should not be none since this is from the non error list. Developer error."
            
            mismatched: bool = surrogate_protospacer_mismatch_single_inference_match_result.value.mismatched
            surrogate_matches: pd.DataFrame = surrogate_protospacer_mismatch_single_inference_match_result.value.surrogate_matches
            protospacer_matches: pd.DataFrame = surrogate_protospacer_mismatch_single_inference_match_result.value.protospacer_matches
            protospacer_surrogate_matches: pd.DataFrame = surrogate_protospacer_mismatch_single_inference_match_result.value.protospacer_surrogate_matches

            # Helper for incrementing nested or flat dicts
            def add_count(counterdict, key, value):
                if contains_sample_barcode:
                    counterdict[cell_barcode][key] += value
                else:
                    counterdict[key] += value

            if mismatched:  # If mismatched, Add count to protospacer/surrogate pair
                assert not surrogate_matches.empty, "Developer error: to be called a mismatch, there must be both separate protospacer and surrogate. No surrogate match (possible no protospacer match)"
                assert not protospacer_matches.empty, "Developer error: to be called a mismatch, there must be both separate protospacer and surrogate. No surrogate match (possible no protospacer match)"
                assert protospacer_surrogate_matches.empty, "Developer error: to be called a mismatch, matches dataframe must be empty."

                for protospacer_matched_whitelist_reporter_series in protospacer_matches.iterrows():
                    for surrogate_matched_whitelist_reporter_series in surrogate_matches.iterrows():
                        dict_index = (tuple(protospacer_matched_whitelist_reporter_series[1]), tuple(surrogate_matched_whitelist_reporter_series[1]))
                        
                        if contains_guide_umi:
                            assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"
                            total_reads = sum(observed_value_counts.values())
                            collapsed = len(observed_value_counts.values())
                            denom = float(protospacer_matches.shape[0] * surrogate_matches.shape[0])

                            add_count(ambiguous_accepted_umi_noncollapsed_mismatch_counterdict, dict_index, total_reads)
                            add_count(ambiguous_accepted_umi_collapsed_mismatch_counterdict, dict_index, collapsed)
                            add_count(ambiguous_spread_umi_noncollapsed_mismatch_counterdict, dict_index, total_reads / denom)
                            add_count(ambiguous_spread_umi_collapsed_mismatch_counterdict, dict_index, collapsed / denom)

                            if (protospacer_matches.shape[0] == 1) and (surrogate_matches.shape[0] == 1):
                                add_count(ambiguous_ignored_umi_noncollapsed_mismatch_counterdict, dict_index, total_reads)
                                add_count(ambiguous_ignored_umi_collapsed_mismatch_counterdict, dict_index, collapsed)
                        else:
                            assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"
                            denom = float(protospacer_matches.shape[0] * surrogate_matches.shape[0])

                            add_count(ambiguous_accepted_mismatch_counterdict, dict_index, observed_value_counts)
                            add_count(ambiguous_spread_mismatch_counterdict, dict_index, observed_value_counts / denom)

                            if (protospacer_matches.shape[0] == 1) and (surrogate_matches.shape[0] == 1):
                                add_count(ambiguous_ignored_mismatch_counterdict, dict_index, observed_value_counts)

            else:  # Matched case
                assert not protospacer_surrogate_matches.empty, f"mismatched==false, but the match dataframe is empty, unexpected paradox. Developer error"
                
                matches = protospacer_surrogate_matches
                for whitelist_reporter_series in matches.iterrows():
                    dict_index = tuple(whitelist_reporter_series[1])
                    if contains_guide_umi:
                        assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"
                        total_reads = sum(observed_value_counts.values())
                        collapsed = len(observed_value_counts.values())
                        denom = float(matches.shape[0])

                        add_count(ambiguous_accepted_umi_noncollapsed_match_counterdict, dict_index, total_reads)
                        add_count(ambiguous_accepted_umi_collapsed_match_counterdict, dict_index, collapsed)
                        add_count(ambiguous_spread_umi_noncollapsed_match_counterdict, dict_index, total_reads / denom)
                        add_count(ambiguous_spread_umi_collapsed_match_counterdict, dict_index, collapsed / denom)

                        if matches.shape[0] == 1:
                            add_count(ambiguous_ignored_umi_noncollapsed_match_counterdict, dict_index, total_reads)
                            add_count(ambiguous_ignored_umi_collapsed_match_counterdict, dict_index, collapsed)
                    else:
                        assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"
                        denom = float(matches.shape[0])
                        add_count(ambiguous_accepted_match_counterdict, dict_index, observed_value_counts)
                        add_count(ambiguous_spread_match_counterdict, dict_index, observed_value_counts / denom)

                        if matches.shape[0] == 1:
                            add_count(ambiguous_ignored_match_counterdict, dict_index, observed_value_counts)

    if contains_sample_barcode:
        for cell_barcode, observed_dict in observed_guide_reporter_umi_counts_inferred.items():
            process_sample(observed_dict, cell_barcode)
    else:
        process_sample(observed_guide_reporter_umi_counts_inferred)

    #
    # Series creation helpers (now aware of sample barcode)
    #
    def create_match_counterseries(counterdict):
        if contains_sample_barcode:
            records = [
                (cell_barcode, *key_tuple, value)
                for cell_barcode, inner in counterdict.items()
                for key_tuple, value in inner.items()
            ]
            columns = ["CellBarcode"] + list(whitelist_guide_reporter_df.columns) + ["value"]
            df = pd.DataFrame.from_records(records, columns=columns)
            return df.set_index(["CellBarcode"] + list(whitelist_guide_reporter_df.columns))["value"]
        else:
            counterseries: pd.Series = whitelist_guide_reporter_df.apply(lambda reporter: counterdict[tuple(reporter)], axis=1)
            counterseries.index = pd.MultiIndex.from_frame(whitelist_guide_reporter_df)
            return counterseries

    def create_mismatch_counterseries(counterdict):
        protospacer_match_suffix = "_ProtospacerMatch"
        surrogate_match_suffix = "_SurrogateMatch"
        whitelist_guide_reporter_df_product = whitelist_guide_reporter_df.merge(whitelist_guide_reporter_df, how='cross', suffixes=(protospacer_match_suffix, surrogate_match_suffix))

        if contains_sample_barcode:
            records = []
            for cell_barcode, inner in counterdict.items():
                for (protospacer_key, surrogate_key), value in inner.items():
                    records.append((cell_barcode, *protospacer_key, *surrogate_key, value))
            columns = ["CellBarcode"] + [c + protospacer_match_suffix for c in whitelist_guide_reporter_df.columns] + [c + surrogate_match_suffix for c in whitelist_guide_reporter_df.columns] + ["value"]
            df = pd.DataFrame.from_records(records, columns=columns)
            return df.set_index(["CellBarcode"] + [c + protospacer_match_suffix for c in whitelist_guide_reporter_df.columns] + [c + surrogate_match_suffix for c in whitelist_guide_reporter_df.columns])["value"]
        else:
            counterseries: pd.Series = whitelist_guide_reporter_df_product.apply(
                lambda pairwise_reporter: counterdict[
                    tuple(pairwise_reporter[[index for index in pairwise_reporter.index if index.endswith(protospacer_match_suffix)]]),
                    tuple(pairwise_reporter[[index for index in pairwise_reporter.index if index.endswith(surrogate_match_suffix)]])
                ],
                axis=1
            )
            counterseries.index = pd.MultiIndex.from_frame(whitelist_guide_reporter_df_product)
            return counterseries


    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results = SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults()
    
    # MATCH counters
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_noncollapsed_match_counterseries = create_match_counterseries(ambiguous_ignored_umi_noncollapsed_match_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_collapsed_match_counterseries = create_match_counterseries(ambiguous_ignored_umi_collapsed_match_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_match_counterseries = create_match_counterseries(ambiguous_ignored_match_counterdict)

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_noncollapsed_match_counterseries = create_match_counterseries(ambiguous_accepted_umi_noncollapsed_match_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_collapsed_match_counterseries = create_match_counterseries(ambiguous_accepted_umi_collapsed_match_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_match_counterseries = create_match_counterseries(ambiguous_accepted_match_counterdict)

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_noncollapsed_match_counterseries = create_match_counterseries(ambiguous_spread_umi_noncollapsed_match_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_collapsed_match_counterseries = create_match_counterseries(ambiguous_spread_umi_collapsed_match_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_match_counterseries = create_match_counterseries(ambiguous_spread_match_counterdict)

    # MISMATCH counters
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_noncollapsed_mismatch_counterseries = create_mismatch_counterseries(ambiguous_ignored_umi_noncollapsed_mismatch_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_collapsed_mismatch_counterseries = create_mismatch_counterseries(ambiguous_ignored_umi_collapsed_mismatch_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_mismatch_counterseries = create_mismatch_counterseries(ambiguous_ignored_mismatch_counterdict)

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_noncollapsed_mismatch_counterseries = create_mismatch_counterseries(ambiguous_accepted_umi_noncollapsed_mismatch_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_collapsed_mismatch_counterseries = create_mismatch_counterseries(ambiguous_accepted_umi_collapsed_mismatch_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_mismatch_counterseries = create_mismatch_counterseries(ambiguous_accepted_mismatch_counterdict)

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_noncollapsed_mismatch_counterseries = create_mismatch_counterseries(ambiguous_spread_umi_noncollapsed_mismatch_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_collapsed_mismatch_counterseries = create_mismatch_counterseries(ambiguous_spread_umi_collapsed_mismatch_counterdict)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_mismatch_counterseries = create_mismatch_counterseries(ambiguous_spread_mismatch_counterdict)

    return surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results


#
# CALLED FUNCTION TO RETRIEVE ALL MATCHSET AND MISMATCHSET COUNTERSERIES
#
@typechecked
def get_counterseries_all_results(observed_guide_reporter_umi_counts_inferred: Union[GeneralMappingInferenceDict, DefaultDict[str, GeneralMappingInferenceDict]], whitelist_guide_reporter_df: pd.DataFrame, contains_guide_barcode: bool, contains_guide_surrogate: bool,  contains_guide_umi: bool, contains_sample_barcode: bool) -> AllMatchSetWhitelistReporterCounterSeriesResults:
    all_match_set_whitelist_reporter_counter_series_results = AllMatchSetWhitelistReporterCounterSeriesResults()
    all_match_set_whitelist_reporter_counter_series_results.protospacer_match = get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_guide_umi, contains_sample_barcode, "protospacer_match")
    if contains_guide_barcode:
        all_match_set_whitelist_reporter_counter_series_results.protospacer_match_barcode_match = get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_guide_umi, contains_sample_barcode, "protospacer_match_barcode_match")
        if contains_guide_surrogate:
            all_match_set_whitelist_reporter_counter_series_results.protospacer_match_surrogate_match_barcode_match = get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_guide_umi, contains_sample_barcode, "protospacer_match_surrogate_match_barcode_match")
            
            all_match_set_whitelist_reporter_counter_series_results.protospacer_mismatch_surrogate_match_barcode_match = get_mismatchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_guide_umi, contains_sample_barcode, "protospacer_mismatch_surrogate_match_barcode_match")
    if contains_guide_surrogate:
        all_match_set_whitelist_reporter_counter_series_results.protospacer_match_surrogate_match = get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_guide_umi, contains_sample_barcode, "protospacer_match_surrogate_match")
        all_match_set_whitelist_reporter_counter_series_results.protospacer_mismatch_surrogate_match = get_mismatchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_guide_umi, contains_sample_barcode, "protospacer_mismatch_surrogate_match")
    
    return all_match_set_whitelist_reporter_counter_series_results



import pysam

def reconstruct_sequence_with_deletions_and_quality(read):
    """Reconstruct the read sequence by adding '-' for deletions and adjusting quality."""
    sequence = []
    qualities = []
    seq_pos = 0  # Position in the original sequence
    max_quality = max(read.query_qualities) if read.query_qualities else 40

    # Iterate through the CIGAR string
    for cigar_op, length in read.cigartuples:
        if cigar_op == 0:  # Match or Mismatch (M)
            sequence.append(read.query_sequence[seq_pos:seq_pos + length])
            qualities.extend(read.query_qualities[seq_pos:seq_pos + length])
            seq_pos += length
        elif cigar_op == 2:  # Deletion (D)
            sequence.append('-' * length)
            qualities.extend([max_quality] * length)  # Use max quality for deletions
        elif cigar_op == 1:  # Insertion (I)
            seq_pos += length  # Skip the inserted bases in the query
        elif cigar_op in {4, 5}:  # Soft or Hard Clipping (S or H)
            if cigar_op == 4:  # Soft clipping affects the query sequence
                seq_pos += length

    return ''.join(sequence), qualities


def bam2fastq_subsetcontig_reconstructindel(input_bam_fn, output_fastq_fn, contig_name = None):
    with pysam.AlignmentFile(input_bam_fn, "rb") as bam, open(output_fastq_fn, "w") as fastq:
        # Iterate through reads mapped to the specified contig
        bam_iterator = None
        if contig_name is None:
            bam_iterator = bam
        else:
            bam_iterator = bam.fetch(contig_name)
        for read in bam_iterator:
            # Skip unmapped reads
            if read.is_unmapped:
                continue

            # Reconstruct sequence with deletions and adjust quality
            reconstructed_sequence, reconstructed_qualities = reconstruct_sequence_with_deletions_and_quality(read)

            if reconstructed_sequence and reconstructed_qualities:
                # Convert quality scores to ASCII characters
                quality_str = ''.join(chr(q + 33) for q in reconstructed_qualities)
                fastq.write(f"@{read.query_name}\n{reconstructed_sequence}\n+\n{quality_str}\n")


import pandas as pd
import numpy as np
import Bio
from Bio import SeqIO
from Bio.Seq import Seq
from pandarallel import pandarallel
import matplotlib.pyplot as plt
from collections import Counter, defaultdict
from typeguard import typechecked
from datetime import date
import re
from multiprocessing import Pool
from functools import partial
from itertools import repeat
import gzip
import random
from enum import Enum
from typing import Callable
from typing import Union, List, Mapping, Tuple, Optional, Any, DefaultDict
from typing import Counter as CounterType
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass

from ..processing import crispr_sequence_encoding
from ..processing import crispr_guide_counting
from ..processing import crispr_sequence_encoding
from ..parsing import reporter_umitools_fastq_parsing, reporter_standard_fastq_parsing
from ..models.mapping_models import WhitelistReporterCountsResult, SampleWhitelistReporterCountsResult, GeneralGuideCountType

# 
# Deprecated 
#
@typechecked
def get_whitelist_reporter_counts_from_umitools_output(whitelist_guide_reporter_df: pd.DataFrame, 
                                                       fastq_r1_fn: str, 
                                                       fastq_r2_fn: str, 
                                                       barcode_pattern_regex: Optional[str] = None, 
                                                       umi_pattern_regex: Optional[str] = None, 
                                                       revcomp_protospacer: bool = False, 
                                                       revcomp_surrogate: bool = True, 
                                                       revcomp_barcode: bool = True, 
                                                       surrogate_hamming_threshold_strict: Optional[int] = 10, 
                                                       barcode_hamming_threshold_strict: Optional[int] = 2, 
                                                       protospacer_hamming_threshold_strict: Optional[int] = 7, 
                                                       cores: int=1) -> WhitelistReporterCountsResult:
    #
    # Get counts of observed FASTQ sequences
    #
    observed_guide_reporter_umi_counts = reporter_umitools_fastq_parsing.get_umitools_observed_sequence_counts(r1_protospacer_fastq_file=fastq_r1_fn, r2_surrogate_fastq_file=fastq_r2_fn, barcode_pattern_regex=barcode_pattern_regex, umi_pattern_regex=umi_pattern_regex, revcomp_protospacer = revcomp_protospacer, revcomp_surrogate = revcomp_surrogate, revcomp_barcode = revcomp_barcode)

    #   
    # Map the observed sequences to the true whitelist sequence
    #
    get_whitelist_reporter_counts_with_umi_PARTIAL = partial(
        crispr_guide_counting.get_whitelist_reporter_counts_with_umi,
        observed_guide_reporter_umi_counts=observed_guide_reporter_umi_counts,
        whitelist_guide_reporter_df=whitelist_guide_reporter_df,
        protospacer_hamming_threshold_strict=protospacer_hamming_threshold_strict,
        surrogate_hamming_threshold_strict=surrogate_hamming_threshold_strict,
        barcode_hamming_threshold_strict=barcode_hamming_threshold_strict,
        cores=cores
    )
    
    if fastq_r2_fn is None: # ONLY R1
        if barcode_pattern_regex is None: # ONLY R1; NO BARCODE
            if umi_pattern_regex is None: # ONLY R1; NO BARCODE; NO UMI
                # These lines are really added just to document the type... observed_guide_reporter_umi_counts is already added to the partial above anyways
                observed_guide_reporter_umi_counts: CounterType[str] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=False, contains_barcode=False, contains_umi = False)
            else: # ONLY R1; NO BARCODE; YES UMI
                observed_guide_reporter_umi_counts: DefaultDict[str, CounterType[str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=False, contains_barcode=False, contains_umi = True)
        else: # ONLY R1; YES BARCODE
            if umi_pattern_regex is None: # ONLY R1; YES BARCODE; NO UMI
                observed_guide_reporter_umi_counts: CounterType[Tuple[str, str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=False, contains_barcode=True, contains_umi = False)
            else: # ONLY R1; YES BARCODE; YES UMI
                observed_guide_reporter_umi_counts: DefaultDict[Tuple[str, str], CounterType[str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=False, contains_barcode=True, contains_umi = True)
    else: # YES R2
        if barcode_pattern_regex is None: # YES R2; NO BARCODE
            if umi_pattern_regex is None: # YES R2; NO BARCODE; NO UMI
                observed_guide_reporter_umi_counts: CounterType[Tuple[str, str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=True, contains_barcode=False, contains_umi = False)
            else: # YES R2; NO BARCODE; YES UMI
                observed_guide_reporter_umi_counts: DefaultDict[Tuple[str, str], CounterType[str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=True, contains_barcode=False, contains_umi = True)
        else: # YES R2; YES BARCODE
            if umi_pattern_regex is None: # YES R2; YES BARCODE; NO UMI
                observed_guide_reporter_umi_counts: CounterType[Tuple[str, str, str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=True, contains_barcode=True, contains_umi = False)
            else: # YES R2; YES BARCODE; YES UMI
                observed_guide_reporter_umi_counts: DefaultDict[Tuple[str, str, str], CounterType[str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=True, contains_barcode=True, contains_umi = True)


# TODO 11/20/2024: Implement UMI parsing via sequence or via R1 header
#@typechecked
def get_whitelist_reporter_counts_from_fastq(whitelist_guide_reporter_df: Optional[pd.DataFrame], 
                                                       fastq_r1_fns: List[str], 
                                                       fastq_r2_fns: Optional[List[str]] = None, 
                                                       
                                                       protospacer_pattern_regex: Optional[str] = None,
                                                       surrogate_pattern_regex: Optional[str] = None,
                                                       guide_barcode_pattern_regex: Optional[str] = None,
                                                       guide_umi_pattern_regex: Optional[str] = None,
                                                       sample_barcode_pattern_regex: Optional[str] = None,

                                                       protospacer_left_flank:Optional[str] = None,
                                                       protospacer_right_flank:Optional[str] = None,
                                                       protospacer_start_position:Optional[int] = None,
                                                       protospacer_end_position:Optional[int] = None,
                                                       protospacer_length: Optional[int] = None,

                                                       surrogate_left_flank:Optional[str] = None,
                                                       surrogate_right_flank:Optional[str] = None,
                                                       surrogate_start_position:Optional[int] = None,
                                                       surrogate_end_position:Optional[int] = None,
                                                       surrogate_length: Optional[int] = None,

                                                       guide_barcode_left_flank:Optional[str] = None,
                                                       guide_barcode_right_flank:Optional[str] = None,
                                                       guide_barcode_start_position:Optional[int] = None,
                                                       guide_barcode_end_position:Optional[int] = None,
                                                       guide_barcode_length: Optional[int] = None,

                                                       guide_umi_left_flank:Optional[str] = None,
                                                       guide_umi_right_flank:Optional[str] = None,
                                                       guide_umi_start_position:Optional[int] = None,
                                                       guide_umi_end_position:Optional[int] = None,
                                                       guide_umi_length: Optional[int] = None,

                                                       sample_barcode_left_flank:Optional[str] = None,
                                                       sample_barcode_right_flank:Optional[str] = None,
                                                       sample_barcode_start_position:Optional[int] = None,
                                                       sample_barcode_end_position:Optional[int] = None,
                                                       sample_barcode_length: Optional[int] = None,

                                                       is_protospacer_r1: Optional[bool] = None, 
                                                       is_surrogate_r1: Optional[bool] = None, 
                                                       is_guide_barcode_r1: Optional[bool] = None,
                                                       is_guide_umi_r1: Optional[bool] = None,
                                                       is_sample_barcode_r1: Optional[bool] = None,

                                                       is_protospacer_header: Optional[bool] = None, 
                                                       is_surrogate_header: Optional[bool] = None, 
                                                       is_guide_barcode_header: Optional[bool] = None,
                                                       is_guide_umi_header: Optional[bool] = None,
                                                       is_sample_barcode_header: Optional[bool] = None,
                                                    
                                                       revcomp_protospacer: Optional[bool] = None, 
                                                       revcomp_surrogate: Optional[bool] = None, 
                                                       revcomp_guide_barcode: Optional[bool] = None, 
                                                       revcomp_guide_umi: Optional[bool] = None,
                                                       revcomp_sample_barcode: Optional[bool] = None, 
                                                        
                                                       surrogate_hamming_threshold_strict: Optional[int] = None, 
                                                       guide_barcode_hamming_threshold_strict: Optional[int] = None, 
                                                       protospacer_hamming_threshold_strict: Optional[int] = None, 

                                                       store_intermediates: bool = False,
                                                       cores: int=1) -> Union[WhitelistReporterCountsResult, SampleWhitelistReporterCountsResult]:
    # Input parameter validation checks

    protospacer_pattern_regex = None if ((protospacer_pattern_regex is not None) and  (protospacer_pattern_regex.strip() == "")) else protospacer_pattern_regex
    surrogate_pattern_regex = None if ((surrogate_pattern_regex is not None) and (surrogate_pattern_regex.strip() == "")) else surrogate_pattern_regex
    guide_barcode_pattern_regex = None if ((guide_barcode_pattern_regex is not None) and  (guide_barcode_pattern_regex.strip() == "")) else guide_barcode_pattern_regex
    guide_umi_pattern_regex = None if ((guide_umi_pattern_regex is not None) and (guide_umi_pattern_regex.strip() == "")) else guide_umi_pattern_regex

    contains_surrogate = (surrogate_pattern_regex is not None) or (surrogate_left_flank is not None) or (surrogate_right_flank is not None) or (surrogate_start_position is not None) or (surrogate_end_position is not None) or (surrogate_length is not None)
    contains_guide_barcode = (guide_barcode_pattern_regex is not None) or (guide_barcode_left_flank is not None) or (guide_barcode_right_flank is not None) or (guide_barcode_start_position is not None) or (guide_barcode_end_position is not None) or (guide_barcode_length is not None)
    contains_guide_umi = (guide_umi_pattern_regex is not None) or (guide_umi_left_flank is not None) or (guide_umi_right_flank is not None) or (guide_umi_start_position is not None) or (guide_barcode_end_position is not None) or (guide_barcode_length is not None)
    contains_sample_barcode = (sample_barcode_pattern_regex is not None) or (sample_barcode_left_flank is not None) or (sample_barcode_right_flank is not None) or (sample_barcode_start_position is not None) or (sample_barcode_end_position is not None) or (sample_barcode_length is not None)
    
    def preprocess_sequence(sequence):
        if sequence is not None:
            sequence = sequence.upper().rstrip()
        return sequence
    
    protospacer_left_flank = preprocess_sequence(protospacer_left_flank)
    protospacer_right_flank = preprocess_sequence(protospacer_right_flank)
    surrogate_left_flank = preprocess_sequence(surrogate_left_flank)
    surrogate_right_flank = preprocess_sequence(surrogate_right_flank)
    guide_barcode_left_flank = preprocess_sequence(guide_barcode_left_flank)
    guide_barcode_right_flank = preprocess_sequence(guide_barcode_right_flank)
    guide_umi_left_flank = preprocess_sequence(guide_umi_left_flank)
    guide_umi_right_flank = preprocess_sequence(guide_umi_right_flank)
    sample_barcode_left_flank = preprocess_sequence(sample_barcode_left_flank)
    sample_barcode_right_flank = preprocess_sequence(sample_barcode_right_flank)

    print(f"Contains surrogate: {contains_surrogate}")
    print(f"Contains guide barcode: {contains_guide_barcode}")
    print(f"Contains guide UMI: {contains_guide_umi}")
    print(f"Contains sample barcode: {contains_sample_barcode}")
    #
    # Get counts of observed FASTQ sequences
    #
    observed_guide_reporter_umi_counts: GeneralGuideCountType = reporter_standard_fastq_parsing.get_standard_observed_sequence_counts(
                                            fastq_r1_fns=fastq_r1_fns, 
                                            fastq_r2_fns=fastq_r2_fns, 
                                            
                                            protospacer_pattern_regex=protospacer_pattern_regex,
                                            surrogate_pattern_regex=surrogate_pattern_regex,
                                            guide_barcode_pattern_regex=guide_barcode_pattern_regex,
                                            guide_umi_pattern_regex=guide_umi_pattern_regex,
                                            sample_barcode_pattern_regex=sample_barcode_pattern_regex,

                                            protospacer_left_flank=protospacer_left_flank,
                                            protospacer_right_flank=protospacer_right_flank,
                                            protospacer_start_position=protospacer_start_position,
                                            protospacer_end_position=protospacer_end_position,
                                            protospacer_length=protospacer_length,

                                            surrogate_left_flank=surrogate_left_flank,
                                            surrogate_right_flank=surrogate_right_flank,
                                            surrogate_start_position=surrogate_start_position,
                                            surrogate_end_position=surrogate_end_position,
                                            surrogate_length=surrogate_length,

                                            guide_barcode_left_flank=guide_barcode_left_flank,
                                            guide_barcode_right_flank=guide_barcode_right_flank,
                                            guide_barcode_start_position=guide_barcode_start_position,
                                            guide_barcode_end_position=guide_barcode_end_position,
                                            guide_barcode_length=guide_barcode_length,

                                            guide_umi_left_flank=guide_umi_left_flank,
                                            guide_umi_right_flank=guide_umi_right_flank,
                                            guide_umi_start_position=guide_umi_start_position,
                                            guide_umi_end_position=guide_umi_end_position,
                                            guide_umi_length=guide_umi_length,

                                            sample_barcode_left_flank=sample_barcode_left_flank,
                                            sample_barcode_right_flank=sample_barcode_right_flank,
                                            sample_barcode_start_position=sample_barcode_start_position,
                                            sample_barcode_end_position=sample_barcode_end_position,
                                            sample_barcode_length=sample_barcode_length,

                                            is_protospacer_r1=is_protospacer_r1, 
                                            is_surrogate_r1=is_surrogate_r1, 
                                            is_guide_barcode_r1=is_guide_barcode_r1,
                                            is_guide_umi_r1=is_guide_umi_r1,
                                            is_sample_barcode_r1=is_sample_barcode_r1,
                                            
                                            is_protospacer_header=is_protospacer_header, 
                                            is_surrogate_header=is_surrogate_header, 
                                            is_guide_barcode_header=is_guide_barcode_header,
                                            is_guide_umi_header=is_guide_umi_header,
                                            is_sample_barcode_header=is_sample_barcode_header,
                                        
                                            revcomp_protospacer=revcomp_protospacer, 
                                            revcomp_surrogate=revcomp_surrogate, 
                                            revcomp_guide_barcode=revcomp_guide_barcode,
                                            revcomp_guide_umi=revcomp_guide_umi,
                                            revcomp_sample_barcode=revcomp_sample_barcode,
                                            
                                            contains_guide_surrogate=contains_surrogate,
                                            contains_guide_barcode=contains_guide_barcode,
                                            contains_guide_umi=contains_guide_umi,
                                            contains_sample_barcode=contains_sample_barcode)

    print(f"Number of unique observed parsed sequences: {len(observed_guide_reporter_umi_counts.keys())}")

    #   
    # Map the observed sequences to the true whitelist sequence
    #
    return crispr_guide_counting.get_whitelist_reporter_counts_with_umi(
        observed_guide_reporter_umi_counts=observed_guide_reporter_umi_counts,
        whitelist_guide_reporter_df=whitelist_guide_reporter_df,
        contains_guide_surrogate=contains_surrogate, 
        contains_guide_barcode=contains_guide_barcode, 
        contains_guide_umi = contains_guide_umi,
        contains_sample_barcode=contains_sample_barcode, 
        protospacer_hamming_threshold_strict=protospacer_hamming_threshold_strict,
        surrogate_hamming_threshold_strict=surrogate_hamming_threshold_strict,
        guide_barcode_hamming_threshold_strict=guide_barcode_hamming_threshold_strict,
        store_intermediates=store_intermediates,
        cores=cores
    )

                




###
### GUIDE PARSING HELPER FUNCTIONS
###
from typeguard import typechecked
from typing import Union, Optional, List, Tuple, DefaultDict
from typing import Counter as CounterType
from ..models.mapping_models import GeneralGuideCountType
from ..models.mapping_models import ProtospacerCounter, ProtospacerDictUMICounter, ProtospacerBarcodeCounter, ProtospacerBarcodeDictUMICounter, ProtospacerSurrogateCounter, ProtospacerSurrogateDictUMICounter, ProtospacerSurrogateBarcodeCounter, ProtospacerSurrogateBarcodeDictUMICounter
from ..models.mapping_models import SampleProtospacerCounter, SampleProtospacerDictUMICounter, SampleProtospacerBarcodeCounter, SampleProtospacerBarcodeDictUMICounter, SampleProtospacerSurrogateCounter, SampleProtospacerSurrogateDictUMICounter, SampleProtospacerSurrogateBarcodeCounter, SampleProtospacerSurrogateBarcodeDictUMICounter
from collections import Counter
from typing import Callable
from functools import partial
from datetime import datetime
import gzip
import re
from Bio.Seq import Seq
from collections import Counter, defaultdict

# This is for grouping the FASTQ lines in 1 tuple.
def grouper(iterable, n=4):
    "s -> (s0,s1,...sn-1), (sn,sn+1,...s2n-1), (s2n,s2n+1,...s3n-1), ..."
    return zip(*[iter(iterable)]*n)

# TODO 20250908 multisample: Likely will change return descriptor to account for sample barcode
@typechecked
def get_standard_observed_sequence_counts(  fastq_r1_fns: List[str], 
                                            fastq_r2_fns: Optional[List[str]], 
                                            
                                            protospacer_pattern_regex: Optional[str],
                                            surrogate_pattern_regex: Optional[str],
                                            guide_barcode_pattern_regex: Optional[str],
                                            guide_umi_pattern_regex: Optional[str],
                                            sample_barcode_pattern_regex: Optional[str],

                                            protospacer_left_flank:Optional[str],
                                            protospacer_right_flank:Optional[str],
                                            protospacer_start_position:Optional[int],
                                            protospacer_end_position:Optional[int],
                                            protospacer_length: Optional[int],

                                            surrogate_left_flank:Optional[str],
                                            surrogate_right_flank:Optional[str],
                                            surrogate_start_position:Optional[int],
                                            surrogate_end_position:Optional[int],
                                            surrogate_length: Optional[int],

                                            guide_barcode_left_flank:Optional[str],
                                            guide_barcode_right_flank:Optional[str],
                                            guide_barcode_start_position:Optional[int],
                                            guide_barcode_end_position:Optional[int],
                                            guide_barcode_length: Optional[int],

                                            guide_umi_left_flank:Optional[str],
                                            guide_umi_right_flank:Optional[str],
                                            guide_umi_start_position:Optional[int],
                                            guide_umi_end_position:Optional[int],
                                            guide_umi_length: Optional[int],

                                            sample_barcode_left_flank:Optional[str],
                                            sample_barcode_right_flank:Optional[str],
                                            sample_barcode_start_position:Optional[int],
                                            sample_barcode_end_position:Optional[int],
                                            sample_barcode_length: Optional[int],


                                            is_protospacer_r1: Optional[bool], 
                                            is_surrogate_r1: Optional[bool], 
                                            is_guide_barcode_r1: Optional[bool],
                                            is_guide_umi_r1: Optional[bool],
                                            is_sample_barcode_r1: Optional[bool],

                                            is_protospacer_header: Optional[bool], 
                                            is_surrogate_header: Optional[bool], 
                                            is_guide_barcode_header: Optional[bool],
                                            is_guide_umi_header: Optional[bool],
                                            is_sample_barcode_header: Optional[bool],
                                        
                                            revcomp_protospacer: Optional[bool], 
                                            revcomp_surrogate: Optional[bool], 
                                            revcomp_guide_barcode: Optional[bool],
                                            revcomp_guide_umi: Optional[bool],
                                            revcomp_sample_barcode: Optional[bool],
                                            contains_guide_surrogate: bool,
                                            contains_guide_barcode: bool,
                                            contains_guide_umi: bool,
                                            contains_sample_barcode: bool,) -> GeneralGuideCountType:
    revcomp = lambda sequence, do_revcomp: str(Seq(sequence).reverse_complement()) if do_revcomp else sequence

    def parse_sequence(
            template_sequence: str,
            sequence_pattern_regex: Optional[str],
            sequence_left_flank:Optional[str],
            sequence_right_flank:Optional[str],
            sequence_start_position:Optional[int],
            sequence_end_position:Optional[int],
            sequence_length: Optional[int],
            revcomp_sequence: Optional[bool],
            sequence_type: str):
        
        if sequence_pattern_regex is not None:
            """
                Parse entirely by regex
            """
            sequence = re.search(sequence_pattern_regex, template_sequence).group(1).rstrip()  
        else:
            """
                Get sequence position start
            """
            if sequence_left_flank is not None:
                """
                    Parse by left flank
                """
                position_start_before_flank = template_sequence.find(sequence_left_flank)
                position_start = position_start_before_flank + len(sequence_left_flank)
                
                if position_start_before_flank == -1:
                    return None # Could not find left_flank, therefore unable to parse
                
            elif sequence_start_position is not None:
                """
                    Use provided position
                """
                position_start = sequence_start_position
            else:
                raise Exception(f"{sequence_type}_left_flank or {sequence_type}_start_position must be provided")
            



            """
                Get sequence position end
            """
            if sequence_right_flank is not None:
                """
                    Parse by left flank
                """
                position_end = template_sequence.find(sequence_right_flank)
                if position_end == -1:
                    return None # Could not find right flank, therefore unable to parse
                
            elif sequence_end_position is not None:
                """
                    Use provided position
                """
                position_end = sequence_end_position
            elif sequence_length is not None:
                """
                    Use provided sequence length
                """
                position_end = position_start + sequence_length
            else:
                raise Exception(f"{sequence_type}_right_flank or {sequence_type}_end_position or {sequence_type}_length must be provided")
            
            sequence = template_sequence[position_start:position_end].upper().rstrip() # rstrip is important, since need to remove newlines if sequence_length exceeds template_sequence length

            if (sequence_length is not None) and (len(sequence) != sequence_length):
                return None # The parsed sequence is not the expected length, so return None
        
        # Return sequence (reverse complement if necessary)
        assert revcomp_sequence is not None, f"revcomp_{sequence_type} must be provided, does the sequence need to be reverse complemented?"
        return revcomp(sequence, revcomp_sequence)

    protospacer_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=protospacer_pattern_regex,
                                                 sequence_left_flank=protospacer_left_flank,
                                                 sequence_right_flank=protospacer_right_flank,
                                                 sequence_start_position=protospacer_start_position,
                                                 sequence_end_position=protospacer_end_position,
                                                 sequence_length=protospacer_length,
                                                 revcomp_sequence=revcomp_protospacer,
                                                 sequence_type="protospacer")
    surrogate_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=surrogate_pattern_regex,
                                                 sequence_left_flank=surrogate_left_flank,
                                                 sequence_right_flank=surrogate_right_flank,
                                                 sequence_start_position=surrogate_start_position,
                                                 sequence_end_position=surrogate_end_position,
                                                 sequence_length=surrogate_length,
                                                 revcomp_sequence=revcomp_surrogate,
                                                 sequence_type="surrogate")
    guide_barcode_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=guide_barcode_pattern_regex,
                                                 sequence_left_flank=guide_barcode_left_flank,
                                                 sequence_right_flank=guide_barcode_right_flank,
                                                 sequence_start_position=guide_barcode_start_position,
                                                 sequence_end_position=guide_barcode_end_position,
                                                 sequence_length=guide_barcode_length,
                                                 revcomp_sequence=revcomp_guide_barcode,
                                                 sequence_type="guide_barcode")
    guide_umi_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=guide_umi_pattern_regex,
                                                 sequence_left_flank=guide_umi_left_flank,
                                                 sequence_right_flank=guide_umi_right_flank,
                                                 sequence_start_position=guide_umi_start_position,
                                                 sequence_end_position=guide_umi_end_position,
                                                 sequence_length=guide_umi_length,
                                                 revcomp_sequence=revcomp_guide_umi,
                                                 sequence_type="guide_umi")
    
    sample_barcode_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=sample_barcode_pattern_regex,
                                                 sequence_left_flank=sample_barcode_left_flank,
                                                 sequence_right_flank=sample_barcode_right_flank,
                                                 sequence_start_position=sample_barcode_start_position,
                                                 sequence_end_position=sample_barcode_end_position,
                                                 sequence_length=sample_barcode_length,
                                                 revcomp_sequence=revcomp_sample_barcode,
                                                 sequence_type="sample_barcode")
    # TODO 09082025 multisample - CREATE sample_barcode parse sequence function.

    def get_sequence_from_read_group(fastq_read_group, parse_sequence_partial, is_r1:bool, is_header: bool, includes_r2: bool, sequence_type: str):
                if includes_r2:
                    r1_header_read, _ = fastq_read_group[0]
                    r1_sequence_read, r2_sequence_read = fastq_read_group[1]
                else:
                    r1_header_read = fastq_read_group[0]
                    r1_sequence_read = fastq_read_group[1]


                if is_r1 is True:
                    sequence = parse_sequence_partial(template_sequence=r1_sequence_read)
                else:
                    if is_header:
                        sequence = parse_sequence_partial(template_sequence=r1_header_read)
                    else:
                        if includes_r2:
                            sequence = parse_sequence_partial(template_sequence=r2_sequence_read)
                        else:
                            raise Exception(f"{sequence_type} cannot be parsed from R1 since is_{sequence_type}_r1 is False, is_{sequence_type}_header is False, and R2 is not provided")

                return sequence
    get_protospacer_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=protospacer_parse_sequence_partial, 
                                                      is_r1=is_protospacer_r1, 
                                                      is_header=is_protospacer_header,
                                                      sequence_type="protospacer")
    get_surrogate_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=surrogate_parse_sequence_partial, 
                                                      is_r1=is_surrogate_r1, 
                                                      is_header=is_surrogate_header,
                                                      sequence_type="surrogate")
    get_guide_barcode_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=guide_barcode_parse_sequence_partial, 
                                                      is_r1=is_guide_barcode_r1, 
                                                      is_header=is_guide_barcode_header,
                                                      sequence_type="guide_barcode")
    get_guide_umi_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=guide_umi_parse_sequence_partial, 
                                                      is_r1=is_guide_umi_r1, 
                                                      is_header=is_guide_umi_header,
                                                      sequence_type="guide_umi")
    get_sample_barcode_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=sample_barcode_parse_sequence_partial, 
                                                      is_r1=is_sample_barcode_r1, 
                                                      is_header=is_sample_barcode_header,
                                                      sequence_type="sample_barcode")
    
    def parse_fastq(fastq_r1_filehandler, fastq_r2_filehandler, sequence_counter: Optional[GeneralGuideCountType] = None):
        NestedDict = lambda: defaultdict(lambda: Counter())

        if fastq_r2_fn is None: # ONLY R1
            
            print("Only R1 FASTQ is provided, R2 NOT provided")
            fastq_single_read_grouper = grouper(fastq_r1_filehandler)
            
            # CONTAINS SAMPLE BARCODE
            if contains_sample_barcode:
                if not contains_guide_barcode: # ONLY R1; NO BARCODE
                    if not contains_guide_umi: # ONLY R1; NO BARCODE; NO UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer-only parsing using function: {get_protospacer_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerCounter = defaultdict(Counter)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer,)][sample_barcode] += 1
                        else:
                            print(f"Performing protospacer+surrrogate parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateCounter = defaultdict(Counter)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (surrogate is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, surrogate)][sample_barcode] += 1

                    else: # ONLY R1; NO BARCODE; YES UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerDictUMICounter = defaultdict(NestedDict)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (umi is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer,)][sample_barcode][umi] += 1
                        else:
                            print(f"Performing protospacer+surrogate+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateDictUMICounter = defaultdict(NestedDict)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (surrogate is not None) and (umi is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, surrogate)][sample_barcode][umi] += 1

                else: # ONLY R1; YES BARCODE
                    if not contains_guide_umi: # ONLY R1; YES BARCODE; NO UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+barcode parsing using protospacer function: {get_protospacer_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerBarcodeCounter = defaultdict(Counter)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (barcode is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, barcode)][sample_barcode] += 1
                        else:
                            print(f"Performing protospacer+surrogate+barcode parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateBarcodeCounter = defaultdict(Counter)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (surrogate is not None) and (barcode is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, surrogate, barcode)][sample_barcode] += 1
                    else: # ONLY R1; YES BARCODE; YES UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+barcode+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerBarcodeDictUMICounter = defaultdict(NestedDict)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (barcode is not None) and (umi is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, barcode)][sample_barcode][umi] += 1
                        else:
                            print(f"Performing protospacer+surrogate+barcode+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateBarcodeDictUMICounter = defaultdict(NestedDict)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (surrogate is not None) and (barcode is not None) and (umi is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, surrogate,barcode)][sample_barcode][umi] += 1
        

            # (Original: no sample barcode)
            elif not contains_sample_barcode:
                if not contains_guide_barcode: # ONLY R1; NO BARCODE
                    if not contains_guide_umi: # ONLY R1; NO BARCODE; NO UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer-only parsing using function: {get_protospacer_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerCounter = Counter()
                            for fastq_single_read_group in fastq_single_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if protospacer is not None:
                                    sequence_counter[(protospacer,)] += 1
                        else:
                            print(f"Performing protospacer+surrrogate parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateCounter = Counter()
                            for fastq_single_read_group in fastq_single_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (surrogate is not None):
                                    sequence_counter[(protospacer, surrogate)] += 1

                    else: # ONLY R1; NO BARCODE; YES UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerDictUMICounter = defaultdict(Counter)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (umi is not None):
                                    sequence_counter[(protospacer,)][umi] += 1
                        else:
                            print(f"Performing protospacer+surrogate+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateDictUMICounter = defaultdict(Counter)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (surrogate is not None) and (umi is not None):
                                    sequence_counter[(protospacer, surrogate)][umi] += 1
                else: # ONLY R1; YES BARCODE
                    if not contains_guide_umi: # ONLY R1; YES BARCODE; NO UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+barcode parsing using protospacer function: {get_protospacer_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerBarcodeCounter = Counter()
                            for fastq_single_read_group in fastq_single_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (barcode is not None):
                                    sequence_counter[(protospacer, barcode)] += 1
                        else:
                            print(f"Performing protospacer+surrogate+barcode parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateBarcodeCounter = Counter()
                            for fastq_single_read_group in fastq_single_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (surrogate is not None) and (barcode is not None):
                                    sequence_counter[(protospacer, surrogate, barcode)] += 1
                    else: # ONLY R1; YES BARCODE; YES UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+barcode+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerBarcodeDictUMICounter = defaultdict(Counter)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (barcode is not None) and (umi is not None):
                                    sequence_counter[(protospacer, barcode)][umi] += 1
                        else:
                            print(f"Performing protospacer+surrogate+barcode+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateBarcodeDictUMICounter = defaultdict(Counter)
                            for fastq_single_read_group in fastq_single_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                                if (protospacer is not None) and (surrogate is not None) and (barcode is not None) and (umi is not None):
                                    sequence_counter[(protospacer, surrogate,barcode)][umi] += 1
        
        else: # YES R2
        
            print("Both R1 FASTQ and R2 FASTQ is provided")
            fastq_paired_read_grouper = grouper(zip(fastq_r1_filehandler, fastq_r2_filehandler))
            
            if contains_sample_barcode:
                if not contains_guide_barcode: # YES R2; NO BARCODE
                    if not contains_guide_umi: # YES R2; NO BARCODE; NO UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer-only parsing using function: {get_protospacer_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerCounter = defaultdict(Counter)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer,)][sample_barcode] += 1
                        else:
                            print(f"Performing protospacer+surrogate parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateCounter = defaultdict(Counter)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (surrogate is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, surrogate)][sample_barcode] += 1
                    else: # YES R2; NO BARCODE; YES UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerDictUMICounter = defaultdict(NestedDict)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (umi is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer,)][sample_barcode][umi] += 1
                        else:
                            print(f"Performing protospacer+surrogate+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateDictUMICounter = defaultdict(NestedDict)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (surrogate is not None) and (umi is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, surrogate)][sample_barcode][umi] += 1
                
                else: # YES R2; YES BARCODE
                    if guide_umi_pattern_regex is None: # YES R2; YES BARCODE; NO UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+barcode parsing using protospacer function: {get_protospacer_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateBarcodeCounter = defaultdict(Counter)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (barcode is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, barcode)][sample_barcode] += 1
                        else:
                            print(f"Performing protospacer+surrogate+barcode parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateBarcodeCounter = defaultdict(Counter)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (surrogate is not None) and (barcode is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, surrogate, barcode)][sample_barcode] += 1
                    else: # YES R2; YES BARCODE; YES UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+barcode+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateBarcodeDictUMICounter = defaultdict(NestedDict)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (barcode is not None) and (umi is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, barcode)][sample_barcode][umi] += 1
                        else:
                            print(f"Performing protospacer+surrogate+barcode+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: SampleProtospacerSurrogateBarcodeDictUMICounter = defaultdict(NestedDict)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                sample_barcode = get_sample_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (surrogate is not None) and (barcode is not None) and (umi is not None) and (sample_barcode is not None):
                                    sequence_counter[(protospacer, surrogate, barcode)][sample_barcode][umi] += 1


            # (Original: no sample barcode)
            if not contains_sample_barcode:
                if not contains_guide_barcode: # YES R2; NO BARCODE
                    if not contains_guide_umi: # YES R2; NO BARCODE; NO UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer-only parsing using function: {get_protospacer_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerCounter = Counter()
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if protospacer is not None:
                                    sequence_counter[(protospacer,)] += 1
                        else:
                            print(f"Performing protospacer+surrrogate parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateCounter = Counter()
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (surrogate is not None):
                                    sequence_counter[(protospacer, surrogate)] += 1
                    else: # YES R2; NO BARCODE; YES UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerDictUMICounter = defaultdict(Counter)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (umi is not None):
                                    sequence_counter[(protospacer,)][umi] += 1
                        else:
                            print(f"Performing protospacer+surrogate+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateDictUMICounter = defaultdict(Counter)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (surrogate is not None) and (umi is not None):
                                    sequence_counter[(protospacer, surrogate)][umi] += 1
                
                else: # YES R2; YES BARCODE
                    if guide_umi_pattern_regex is None: # YES R2; YES BARCODE; NO UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+barcode parsing using protospacer function: {get_protospacer_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateBarcodeCounter = Counter()
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (barcode is not None):
                                    sequence_counter[(protospacer, barcode)] += 1
                        else:
                            print(f"Performing protospacer+surrogate+barcode parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateBarcodeCounter = Counter()
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (surrogate is not None) and (barcode is not None):
                                    sequence_counter[(protospacer, surrogate, barcode)] += 1
                    else: # YES R2; YES BARCODE; YES UMI
                        if not contains_guide_surrogate:
                            print(f"Performing protospacer+barcode+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateBarcodeDictUMICounter = defaultdict(Counter)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (barcode is not None) and (umi is not None):
                                    sequence_counter[(protospacer, barcode)][umi] += 1
                        else:
                            print(f"Performing protospacer+surrogate+barcode+UMI parsing using protospacer function: {get_protospacer_from_read_group_partial} \nsurrogate function: {get_surrogate_from_read_group_partial} \nbarcode function: {get_guide_barcode_from_read_group_partial} \nUMI function: {get_guide_umi_from_read_group_partial}")
                            if sequence_counter is None:
                                sequence_counter: ProtospacerSurrogateBarcodeDictUMICounter = defaultdict(Counter)
                            for fastq_paired_read_group in fastq_paired_read_grouper:
                                protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                barcode = get_guide_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                umi = get_guide_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                                if (protospacer is not None) and (surrogate is not None) and (barcode is not None) and (umi is not None):
                                    sequence_counter[(protospacer, surrogate, barcode)][umi] += 1
        
        return sequence_counter
    


    # Iterate through FASTQs
    sequence_counter: Optional[GeneralGuideCountType] = None
    for fastq_index in range(len(fastq_r1_fns)):
        print(f"Processing FASTQ {fastq_index} of {len(fastq_r1_fns)}")

        before_file_loading_time = datetime.now()
        fastq_r1_fn = fastq_r1_fns[fastq_index]
        
        # Open R1 file handler
        fastq_r1_filehandler = None
        if fastq_r1_fn.endswith('.gz'):
            print(f"Opening FASTQ.gz file with gzip, filename={fastq_r1_fn}")
            fastq_r1_filehandler = gzip.open(fastq_r1_fn, "rt", encoding="utf-8")
        else:
            print(f"Opening FASTQ file, filename={fastq_r1_fn}")
            fastq_r1_filehandler = open(fastq_r1_fn, "r")
        
        # Open R2 file handler if provided
        fastq_r2_filehandler = None
        if fastq_r2_fns is not None:
            fastq_r2_fn = fastq_r2_fns[fastq_index]

            if fastq_r2_fn.endswith('.gz'):
                print(f"Opening FASTQ.gz file with gzip, filename={fastq_r2_fn}")
                fastq_r2_filehandler = gzip.open(fastq_r2_fn, "rt", encoding="utf-8")
            else:
                print(f"Opening FASTQ file, filename={fastq_r2_fn}")
                fastq_r2_filehandler = open(fastq_r2_fn, "r")
        after_file_loading_time = datetime.now()

        print(f"{(after_file_loading_time-before_file_loading_time).seconds} seconds for file loading")
        sequence_counter = parse_fastq(fastq_r1_filehandler, fastq_r2_filehandler, sequence_counter)
        after_parsing_time = datetime.now()
        print(f"{(after_parsing_time-after_file_loading_time).seconds} seconds for parsing")

        # Close the file handlers when done
        if fastq_r1_filehandler is not None:
            fastq_r1_filehandler.close()
        if fastq_r2_filehandler is not None:
            fastq_r2_filehandler.close()

    return sequence_counter


import numpy as np

# Vector encoding of each base, since we want to vectorized.
# This is for encoding the whitelist library
non_ambigious_encoding_dict_whitelist = dict({
    "A": np.asarray([1,0,0,0]),
    "C": np.asarray([0,1,0,0]),
    "G": np.asarray([0,0,1,0]),
    "T": np.asarray([0,0,0,1]),
    "U": np.asarray([0,0,0,1]),
    "-": np.asarray([0,0,0,0]), # NOTE: This representes a deletion token
    "X": np.asarray([0,0,0,0]) # NOTE: This represents a padding token for sequences. Set to some high number, so that the calculated hamming distance is very high and won't be selected (which will happen if the observed sequence is greater than the whitelist sequence) # Correct 20250106 should be set to 0, so that it will just be a hamming increment of 1)
})

# Encoding of observed sequences
# Set mismatches to 3 so that after bitflip, hamming penalty is 2
non_ambigious_encoding_dict_observed = dict({
    "A": np.asarray([1,3,3,3]),
    "C": np.asarray([3,1,3,3]),
    "G": np.asarray([3,3,1,3]),
    "T": np.asarray([3,3,3,1]),
    "U": np.asarray([3,3,3,1]),
    "-": np.asarray([3,3,3,3]), 
    "X": np.asarray([3,3,3,3])
})

'''
 Helper function for retrieving the encoding for ambigious bases based on IUPAC codes
'''
def ambiguity_encoder(bases):
    return np.logical_or.reduce([non_ambigious_encoding_dict_whitelist[base] for base in bases]).astype(int)

'''
    Final dictionary for getting encoding of each IUPAC base
'''
# Final encoding of whitelist
full_encoding_dict_whitelist = dict({
    "A": non_ambigious_encoding_dict_whitelist["A"],
    "C": non_ambigious_encoding_dict_whitelist["C"],
    "G": non_ambigious_encoding_dict_whitelist["G"],
    "T": non_ambigious_encoding_dict_whitelist["T"], 
    "X": non_ambigious_encoding_dict_whitelist["X"], # NOTE: This represents a padding token for sequences
    "-": non_ambigious_encoding_dict_whitelist["-"], # NOTE: This represents a padding token for sequences 
    "R": ambiguity_encoder(["A", "G"]),
    "Y": ambiguity_encoder(["C", "T"]),
    "S": ambiguity_encoder(["G", "C"]),
    "W": ambiguity_encoder(["A", "T"]),
    "K": ambiguity_encoder(["G", "T"]),
    "M": ambiguity_encoder(["A", "C"]),
    "B": ambiguity_encoder(["C", "G", "T"]),
    "D": ambiguity_encoder(["A", "G", "T"]),
    "H": ambiguity_encoder(["A", "C", "T"]),
    "V": ambiguity_encoder(["A", "C", "G"]),
    "N": ambiguity_encoder(["A", "C", "G", "T"]),
})

full_encoding_dict_observed = dict({
    "A": non_ambigious_encoding_dict_observed["A"],
    "C": non_ambigious_encoding_dict_observed["C"],
    "G": non_ambigious_encoding_dict_observed["G"],
    "T": non_ambigious_encoding_dict_observed["T"], 
    "X": non_ambigious_encoding_dict_observed["X"], # NOTE: This represents a padding token for sequences
    "-": non_ambigious_encoding_dict_observed["-"], # NOTE: This represents a padding token for sequences 
    "R": (ambiguity_encoder(["A", "G"])^1)*3,
    "Y": (ambiguity_encoder(["C", "T"])^1)*3,
    "S": (ambiguity_encoder(["G", "C"])^1)*3,
    "W": (ambiguity_encoder(["A", "T"])^1)*3,
    "K": (ambiguity_encoder(["G", "T"])^1)*3,
    "M": (ambiguity_encoder(["A", "C"])^1)*3,
    "B": (ambiguity_encoder(["C", "G", "T"])^1)*3,
    "D": (ambiguity_encoder(["A", "G", "T"])^1)*3,
    "H": (ambiguity_encoder(["A", "C", "T"])^1)*3,
    "V": (ambiguity_encoder(["A", "C", "G"])^1)*3,
    "N": (ambiguity_encoder(["A", "C", "G", "T"])^1)*3,
})

'''
    Main function to encode a single base
'''
def encode_DNA_base_whitelist(char):
    return full_encoding_dict_whitelist[char]
encode_DNA_base_whitelist_vectorized = np.vectorize(encode_DNA_base_whitelist, signature='()->(n)') # Vectorized function for a string (i.e. gRNA)

def encode_DNA_base_observed(char):
    return full_encoding_dict_observed[char]
encode_DNA_base_observed_vectorized = np.vectorize(encode_DNA_base_observed, signature='()->(n)') # Vectorized function for a string (i.e. gRNA)

'''
    Function for converting string (i.e. gRNA) into a np array of chars  - may be deprecated (NOTE 20221202)
'''
def numpify_string(string):
    return np.array(list(string), dtype=str)
numpify_string_vectorized = np.vectorize(numpify_string, signature='()->(n)') # Vectorize the function

def encode_guide_series_whitelist(guide_series) -> np.array:
    guide_numpy = guide_series.to_numpy(dtype=object)
    guide_numpy = guide_numpy.astype(str)
    guide_numpy_char = np.array(list(map(list, guide_numpy))) # Map into a list of list of characters
    guide_numpy_encoding = encode_DNA_base_whitelist_vectorized(guide_numpy_char)
    return guide_numpy_encoding

def encode_guide_series_observed(guide_series) -> np.array:
    guide_numpy = guide_series.to_numpy(dtype=object)
    guide_numpy = guide_numpy.astype(str)
    guide_numpy_char = np.array(list(map(list, guide_numpy))) # Map into a list of list of characters
    guide_numpy_encoding = encode_DNA_base_observed_vectorized(guide_numpy_char)
    return guide_numpy_encoding

def retrieve_hamming_distance_whitelist(observed_guide_encoded, whitelist_guide_encoded):
    '''
        This takes a encoded guide sequence and a list of encoded whitelisted guides and matrix computes the hamming distance of the 
        encoded guide across all whitelisted guides in a single operation.
        
        
        (observed_guide_encoded*whitelist_guide_encoded[:, np.newaxis]).clip(0, 2).sum(axis=3) # Determines 

        # observed_guide_encoded*whitelist_guide_encoded[:, np.newaxis] performs multiplication of the target encoding with every whitelist encoding
        #
        
    '''
    # Compute the per-character un-bit flipped hamming penalty 
    encoding_product = (observed_guide_encoded*whitelist_guide_encoded[:, np.newaxis]).clip(0,3)
    # Need to get the minimum positive un-bit flipped hamming penalty (so we are removing 0s)
    masked_encoding_product = np.where(encoding_product > 0, encoding_product, np.inf)
    # Get the minimum positive un-bit flipped hamming penalty for each comparison (axis=-1)
    min_positive = np.min(masked_encoding_product, axis=-1)
    # If there is no un-bit flipped hamming penalty (all 0s, or all np.inf after masking), set the minimum positive as 0 
    min_positive = np.where(np.isinf(min_positive), 0, min_positive).astype(int)
    # Bit-flip to get the hamming distance, take the sum to get the per-encoding comparison hamming distance, and divide by 2 since the hamming distance is * 2 since working in bits. 
    hamming_distance = (min_positive^1).sum(axis=2).flatten()/2
    return hamming_distance
